/*
 * Copyright (c) 2001, 2002, 2004 Adaptec Inc.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions, and the following disclaimer,
 *    without modification.
 * 2. Redistributions in binary form must reproduce at minimum a disclaimer
 *    substantially similar to the "NO WARRANTY" disclaimer below
 *    ("Disclaimer") and any redistribution must be conditioned upon
 *    including a substantially similar Disclaimer requirement for further
 *    binary redistribution.
 * 3. Neither the names of the above-listed copyright holders nor the names
 *    of any contributors may be used to endorse or promote products derived
 *    from this software without specific prior written permission.
 *
 * Alternatively, this software may be distributed under the terms of the
 * GNU General Public License ("GPL") version 2 as published by the Free
 * Software Foundation.
 *
 * NO WARRANTY
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTIBILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 * HOLDERS OR CONTRIBUTORS BE LIABLE FOR SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
 * STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING
 * IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGES.
 */

#define __packed __attribute__((packed))

#define DDF_DEFAULT_PRIORITY 0x8000
#define DDF_BOOT_PRIORITY    0

#define DDF_GUID_BYTES		24

/* Definitions from the DDF spec, rev 0.8 */
#define	DDF_HEADER_BYTES	512
#define DDF_HEADER_LENGTH	1
struct ddf_header {
	uint32_t	Signature;
#define	DDF_HEADER_SIGNATURE	0xde11de11
	uint32_t	CRC;
	uint8_t		DDF_Header_GUID[DDF_GUID_BYTES];
	uint8_t		DDF_rev[8];
	uint32_t	Sequence_Number;
	uint32_t	TimeStamp;
	uint8_t		Open_Flag;
#define	DDF_HEADER_CLOSE	0x00
#define	DDF_HEADER_OPEN_MASK	0x0f
#define	DDF_HEADER_OPEN_ANCHOR	0xff
	uint8_t		Foreign_Flag;
	uint8_t		Diskgrouping;
	uint8_t		rsrvd1[13];
	uint8_t		Header_ext[32];
	uint64_t	Primary_Header_LBA;
	uint64_t	Secondary_Header_LBA;
	uint8_t		Header_Type;
#define	DDF_HEADER_ANCHOR	0x00
#define	DDF_HEADER_PRIMARY	0x01
#define	DDF_HEADER_SECONDARY	0x02
	uint8_t		rsrvd2[3];
	uint32_t	WorkSpace_Length;
	uint64_t	WorkSpace_LBA;
	uint16_t	Max_PD_Entries;
	uint16_t	Max_VD_Entries;
	uint16_t	Max_Partitions;
	uint16_t	Configuration_Record_Length;
	uint8_t		rsrvd3[56];
	uint32_t	cd_section;	/* Controller_Data_Section */
	uint32_t	cd_length;	/* Controller_Data_Section_Length */
	uint32_t	pdr_section;	/* Physical_Drive_Records_Section */
	uint32_t	pdr_length;	/* Physical_Drive_Records_Length */
	uint32_t	vdr_section;	/* Virtual_Drive_Records_Section */
	uint32_t	vdr_length;	/* Virtual_Drive_Records_Length */
	uint32_t	cr_section;	/* Configuration_Records_Section */
	uint32_t	cr_length;	/* Configuration_Records_Length */
	uint32_t	pdd_section;	/* Physical_Drive_Data_Section */
	uint32_t	pdd_length;	/* Physical_Drive_Data_Length */
	uint32_t	bbmlog_section;	/* BBM_Log_Section */
	uint32_t	bbmlog_length;	/* BBM_Log_Section_Length */
	uint32_t	Diagnostic_Space;
	uint32_t	Diagnostic_Space_Length;
	uint32_t	vendor_section;
	uint32_t	vendor_length;
} __packed;

#define DDF_ADPT_CD_OFFSET	1
#define DDF_ADPT_CD_LENGTH	1
struct ddf_cd_record {
	uint32_t	Signature;
#define	DDF_CD_SIGNATURE	0xad111111
	uint32_t	CRC;
	uint8_t		Controller_GUID[DDF_GUID_BYTES];
	struct {
		uint16_t	Vendor_ID;
		uint16_t	Device_ID;
		uint16_t	SubVendor_ID;
		uint16_t	SubDevice_ID;
	} Adapter_Type __packed;
	uint8_t		NID[24];
	uint8_t		Controller_Data[448];
} __packed;

union ddf_device {
	struct {
		uint8_t		Lun;
		uint8_t		Id;
	} scsi __packed;
	uint16_t	port;
};

struct ddf_pathinfo {
	union ddf_device	device;
	uint8_t			channel;
	uint8_t			type;
#define	DDF_PATH_UNKNOWN	(1 << 0)
#define	DDF_PATH_SCSI		(1 << 1)
#define	DDF_PATH_SAS		(1 << 2)
#define	DDF_PATH_SATA		(1 << 3)
#define	DDF_PATH_FC		(1 << 4)
} __packed;

struct ddf_pd_entry {
	uint8_t			PD_GUID[DDF_GUID_BYTES];
	uint32_t		PD_Reference;
	uint16_t		PD_Type;
#define	DDF_PDE_GUID_FORCE	(1 << 0)
#define	DDF_PDE_PARTICIPATING	(1 << 1)
#define	DDF_PDE_GLOBAL_SPARE	(1 << 2)
#define	DDF_PDE_CONFIG_SPARE	(1 << 3)
#define	DDF_PDE_FOREIGN		(1 << 4)
#define	DDF_PDE_LEGACY		(1 << 5)
	uint16_t		PD_State;
#define	DDF_PDE_ONLINE		(1 << 0)
#define	DDF_PDE_FAILED		(1 << 1)
#define	DDF_PDE_REBUILD		(1 << 2)
#define	DDF_PDE_TRANSITION	(1 << 3)
#define	DDF_PDE_PFA		(1 << 4)
#define	DDF_PDE_UNRECOVERED	(1 << 5)
#define	DDF_PDE_MISSING		(1 << 6)
	uint64_t		Configured_Size;
	struct ddf_pathinfo	Path_0;
	struct ddf_pathinfo	Path_1;
	uint8_t			rsrvd2[16];
} __packed;

#define DDF_ADPT_PDR_OFFSET	2
#define DDF_ADPT_PDR_LENGTH	2
struct ddf_pd_record {
	uint32_t		Signature;
#define	DDF_PDR_SIGNATURE	0x22222222
	uint32_t		CRC;
	uint16_t		Populated_PDEs;
	uint16_t		Max_PD_Entries;
	uint8_t			rsrvd1[52];
	struct ddf_pd_entry	entry[0];
} __packed;

struct ddf_vd_entry {
	uint8_t			VD_GUID[DDF_GUID_BYTES];
	uint16_t		VD_Number;
	uint8_t			rsrvd1[2];
	uint16_t		VD_Type;
#define DDF_VDE_SHARED		(1 << 0)
#define	DDF_VDE_ENFORCE_GROUP	(1 << 1)
#define	DDF_VDE_OWNER_ID_VALID	(1 << 3)
	union ddf_device	Device;
	uint8_t			VD_State;
#define	DDF_VDE_OPTIMAL		0x00
#define	DDF_VDE_DEGRADED	0x01
#define	DDF_VDE_DELETED		0x02
#define	DDF_VDE_MISSING		0x03
#define	DDF_VDE_FAILED		0x04
#define	DDF_VDE_STATE_MASK	0x07
#define	DDF_VDE_MORPH		(1 << 3)
#define	DDF_VDE_DIRTY		(1 << 4)
	uint8_t			Init_State;
#define	DDF_VDE_UNINTIALIZED	0x00
#define	DDF_VDE_INIT_QUICK	0x01
#define	DDF_VDE_INIT_FULL	0x02
#define	DDF_VDE_INIT_MASK	0x03
#define	DDF_VDE_UACCESS_NORM	0x00
#define	DDF_VDE_UACCESS_RO	0x80
#define	DDF_VDE_UACCESS_BLOCKED	0xc0
#define	DDF_VDE_UACCESS_MASK	0xc0
	uint8_t			rsrvd2[14];
	uint8_t			VD_Name[16];
} __packed;

#define DDF_ADPT_VDR_OFFSET	4
#define DDF_ADPT_VDR_LENGTH	1
struct ddf_vd_record {
	uint32_t		Signature;
#define	DDF_VD_RECORD_SIGNATURE	0xdddddddd
	uint32_t		CRC;
	uint16_t		Populated_VDEs;
	uint16_t		Max_VD_Entries;
	uint8_t			rsrvd1[52];
	struct ddf_vd_entry	entry[0];
} __packed;

#define DDF_ADPT_VDCR_OFFSET	5
#define DDF_ADPT_VDCR_LENGTH	2
#define	DDF_VDCR_PDS_OFFSET	512
struct ddf_vdc_record {
	uint32_t	Signature;
#define	DDF_VDCR_SIGNATURE	0xeeeeeeee
	uint32_t	CRC;
	uint8_t		VD_GUID[DDF_GUID_BYTES];
	uint32_t	Timestamp;
	uint32_t	Sequence_Number;
	uint8_t		rsrvd1[DDF_GUID_BYTES];
	uint16_t	Primary_Element_Count;
	uint8_t		Stripe_Size;
	uint8_t		Primary_RAID_Level;
	uint8_t		RLQ;
	uint8_t		Secondary_Element_Count;
	uint8_t		Secondary_Element_Seq;
	uint8_t		Secondary_RAID_Level;
	uint64_t	Block_Count;
	uint64_t	Size;
	uint8_t		rsrvd2[8];
	uint32_t	Associated_Spares[8];
	uint64_t	Cache_Flags;
#define DDF_VDCR_CACHE_WB		(1 << 0)
#define	DDF_VDCR_CACHE_WB_ADAPTIVE	(1 << 1)
#define	DDF_VDCR_CACHE_RA		(1 << 2)
#define	DDF_VDCR_CACHE_RA_ADAPTIVE	(1 << 3)
#define	DDF_VDCR_CACHE_WCACHE_NOBATTERY	(1 << 4)
#define	DDF_VDCR_CACHE_WCACHE_ALLOW	(1 << 5)
#define	DDF_VDCR_CACHE_RCACHE_ALLOW	(1 << 6)
#define	DDF_VDCR_CACHE_VENDOR		(1 << 7)
	uint8_t		BG_Rate;
	uint8_t		rsrvd4[3];
	uint8_t		rsrvd5[52];
	uint8_t		rsrvd6[192];
	uint8_t		V0[32];
	uint8_t		V1[32];
	uint8_t		V2[16];
	uint8_t		V3[16];
	uint8_t		Vendor_Scratch[32];
} __packed;

#define DDF_ADPT_VUCR_OFFSET	5
#define DDF_ADPT_VUCR_LENGTH	1
struct ddf_vuc_record {
	uint32_t	Signature;
#define	DDF_VUCR_SIGNATURE	0x88888888
	uint32_t	CRC;
	uint8_t		VD_GUID[DDF_GUID_BYTES];
} __packed;

struct ddf_sa_entry {
	uint8_t		VD_GUID[DDF_GUID_BYTES];
	uint16_t	Secondary_Element;
	uint8_t		rsrvd2[6];
} __packed;

#define DDF_ADPT_SA_OFFSET	5
#define DDF_ADPT_SA_LENGTH	1
struct ddf_sa_record {
	uint32_t	Signature;
#define	DDF_SA_SIGNATURE	0x55555555
	uint32_t	CRC;
	uint32_t	Timestamp;
	uint32_t	Channel_Affinity;
#define	DDF_SAR_CHANNEL_AFFINITY	0x01
	uint8_t		rsrvd1[3];
	uint8_t		Spare_Type;
#define	DDF_SAR_TYPE_DEDICATED		(1 << 0)
#define	DDF_SAR_TYPE_REVERTIBLE		(1 << 1)
	uint8_t		rsrvd2[12];
	struct ddf_sa_entry	entry[0];
} __packed;

#define DDF_ADPT_PDD_OFFSET	9
#define DDF_ADPT_PDD_LENGTH	1
struct ddf_pdd_record {
	uint32_t	Signature;
#define	DDF_PDD_SIGNATURE	0x33333333
	uint32_t	CRC;
	uint8_t		PD_GUID[DDF_GUID_BYTES];
	uint32_t	PD_Reference;
	uint8_t		Forced_Ref_Flag;
#define	DDF_PDD_FORCED_REF	0x01
	uint8_t		Forced_PD_GUID_Flag;
#define	DDF_PDD_FORCED_GUID	0x01
	uint8_t		Vendor_Scratch[32];
	uint8_t		rsrvd2[442];
} __packed;

struct ddf_bbm_entry {
	uint64_t	Defective_Block_Start;
	uint32_t	Spare_Block_Offset;
	uint16_t	Remapped_Count;
	uint16_t	rsrvd[2];
};

struct ddf_bbm_log {
	uint32_t	Signature;
#define	DDF_BBML_SIGNATURE	0xabadb10c
	uint32_t	CRC;
	uint16_t	Entry_Count;
	uint16_t	Spare_Block_Count;
	uint8_t		rsrvd1[4];
	uint64_t	First_Spare_LBA;
	uint64_t	Mapped_Block_Entry[0];
} __packed;

#define DDF_ADPT_VENDOR_OFFSET	10
#define DDF_ADPT_VENDOR_LENGTH	1
struct ddf_vendor_log {
	uint32_t	Signature;
#define	DDF_VENDOR_LOG_SIGNATURE	0x01dbeef0
	uint32_t	CRC;
	uint8_t		Log_Owner[8];
	uint8_t		data[0];
} __packed;

struct adpt_vendor_data {
	uint8_t		Reserved[16];
	uint32_t	CheckpointLBA;
	uint32_t	VerifyDate;
	uint32_t	RecreateDate;
	uint16_t	MemberIndex;
	uint8_t		PhysDriveIndex;
	uint8_t		CheckpointActivity;
#define DDF_ADPT_CHKPNT_ACT_NONE	0
#define DDF_ADPT_CHKPNT_ACT_VERIFY	1
#define DDF_ADPT_CHKPNT_ACT_REBUILD	2
#define DDF_ADPT_CHKPNT_ACT_ZERO	3
	uint8_t		SpareCnt;
	uint8_t		irocFlags;
#define DDF_ADPT_IF_VERIFY_WITH_AUTOFIX	0x01
#define DDF_ADPT_IF_BOOT_VOLUME		0x80
	uint8_t		RedundancySubstrate;
#define DDF_ADPT_LRRSS_OK	0
#define DDF_ADPT_LRRSS_FAILED	1
#define DDF_ADPT_LRRSS_RECOVERY	2
	uint8_t		Dirty;
	uint8_t		rsrvd[460];
} __packed;
