/*
   md.h : Multiple Devices driver for Linux
          Copyright (C) 1996-98 Ingo Molnar, Gadi Oxman
          Copyright (C) 1994-96 Marc ZYNGIER
	  <zyngier@ufr-info-p7.ibp.fr> or
	  <maz@gloups.fdn.fr>
	  
   Pluggable metadata support.
   	  Copyright (C) 2004 Adaptec Inc.

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2, or (at your option)
   any later version.
   
   You should have received a copy of the GNU General Public License
   (for example /usr/src/linux/COPYING); if not, write to the Free
   Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.  
*/

#ifndef _MD_H
#define _MD_H

#include <linux/blkdev.h>
#include <asm/semaphore.h>
#include <linux/major.h>
#include <linux/ioctl.h>
#include <linux/types.h>
#include <linux/bitops.h>
#include <linux/module.h>
#include <linux/hdreg.h>
#include <linux/proc_fs.h>
#include <linux/smp_lock.h>
#include <linux/delay.h>
#include <net/checksum.h>
#include <linux/random.h>
#include <linux/kernel_stat.h>
#include <asm/io.h>
#include <linux/completion.h>
#if LINUX_VERSION_CODE <= KERNEL_VERSION(2,4,9)
#include "mempool.h"
#else
#include <linux/mempool.h>
#endif
#include <linux/list.h>
#include <linux/reboot.h>
#include <linux/vmalloc.h>
#include <linux/blkpg.h>
#include <linux/pagemap.h>
#include <linux/time.h>

#ifndef TRUE
#define TRUE 1
#endif

#ifndef FALSE
#define FALSE 0
#endif

typedef unsigned long sector_t;
#define BDEVNAME_SIZE 32

#if LINUX_VERSION_CODE <= KERNEL_VERSION(2,4,9)
#define container_of(ptr, type, member) ({			\
	const typeof( ((type *)0)->member ) *__mptr = (ptr);	\
	(type *)( (char *)__mptr - offsetof(type,member) );})

#define list_for_each_entry(pos, head, member)				\
	for (pos = list_entry((head)->next, typeof(*pos), member),	\
		prefetch(pos->member.next);				\
	&pos->member != (head);						\
	pos = list_entry(pos->member.next, typeof(*pos), member),	\
		prefetch(pos->member.next))
#endif

#if LINUX_VERSION_CODE <= KERNEL_VERSION(2,4,9)
static inline void md_flush_signals (void)
{
	spin_lock(&current->sigmask_lock);
	flush_signals(current);
	spin_unlock(&current->sigmask_lock);
}
#else
static inline void md_flush_signals (void)
{
	spin_lock(&current->sighand->siglock);
	flush_signals(current);
	spin_unlock(&current->sighand->siglock);
}
#endif

/*
 * 'md_p.h' holds the 'physical' layout of RAID devices
 * 'md_u.h' holds the user <=> kernel API
 *
 * 'md_k.h' holds kernel internal definitions
 */

#include <linux/raid/emd/md_u.h>
#include <linux/raid/emd/md_p.h>
#include <linux/raid/emd/md_k.h>

/*
 * Different major versions are not compatible.
 * Different minor versions are only downward compatible.
 * Different patchlevel versions are downward and upward compatible.
 */
#define MD_MAJOR_VERSION                0 
#define MD_MINOR_VERSION                21 
#define MD_PATCHLEVEL_VERSION           0

extern int register_md_personality (int p_num, mdk_personality_t *p);
extern int unregister_md_personality (int p_num);
extern int register_md_metapersonality (mdk_metapersonality_t *p);
extern int unregister_md_metapersonality (mdk_metapersonality_t *p);
extern mdk_thread_t * md_register_thread (mdk_thread_handler_t *run,
				mddev_t *mddev, void *arg, const char *name);
extern void md_unregister_thread (mdk_thread_t *thread);
extern void md_update_thread_string(mdk_thread_t *thread, const char *fmt);
extern void md_wakeup_thread(mdk_thread_t *thread);
extern void md_interrupt_thread (mdk_thread_t *thread);
extern void md_handle_safemode(mddev_t *mddev);
extern void md_sync_acct(mdk_member_t *member, unsigned long nr_sectors);

extern void md_print_devices (void);
extern void md_get_array_stats(mddev_t *array, int *optimal, int *degraded,
			       int *failed, int *build_targets);
extern int md_read_blocks (kdev_t dev, sector_t start_sector,
			   int num_sectors, uint8_t *dest_buf);
extern int md_write_blocks (kdev_t dev, sector_t start_sector,
			    int num_sectors, uint8_t *src_buf);
extern mdk_rdev_t *md_rdev_alloc(mdk_member_t *parent, mdk_msubtype_t subtype,
				 dev_t newdev, int priority);
extern mddev_t *md_mddev_alloc(mdk_member_t *parent, mdk_msubtype_t subtype,
			       int priority);
extern int md_member_add_child(mdk_member_t *child, mdk_member_t *parent);
extern void md_member_orphan(mdk_member_t *member);
extern mdk_personality_t *md_find_pers(int level);
extern mdk_personality_t *md_pers_get(mdk_personality_t *);
extern void md_pers_put(mdk_personality_t *);
extern mdk_metapersonality_t *md_metapers_get(mdk_metapersonality_t *);
extern void md_metapers_put(mdk_metapersonality_t *);
extern struct block_device *md_get_blkdev(dev_t dev);
extern void md_release_blkdev(struct block_device *blkdev);
extern void kick_rdev_from_array(mdk_rdev_t * rdev);
extern int md_check_chunksize(mddev_t *mddev);
extern int md_assign_unit_number(mddev_t *mddev);
extern void md_teardown_member(mdk_member_t *member);
extern int md_export_array(mddev_t *mddev);
extern int md_init_workitem_support(mddev_t *mddev,
				    mdk_workitem_support_t *,
				    const char *);
extern void md_teardown_workitem_support(mdk_workitem_support_t *workinfo);
extern int md_init_workitem(mdk_workitem_t *work, u_int opcode,
			    mdk_member_callback_t *control,
			    mdk_work_handler_t *handler, void *data);

extern void md_cancel_workitems(mdk_workitem_support_t *workinfo);
extern void md_wait_workitem(mdk_workitem_support_t *, mdk_workitem_t *);
extern mdk_workitem_flags_t md_find_workitem(mdk_workitem_support_t *workinfo,
					     u_int opcode, int *found);
extern int md_queue_workitem(mdk_workitem_support_t *workinfo,
			     mdk_workitem_t *work, int head_of_queue);
extern void md_exit_safemode(mddev_t *mddev, mdk_sync_support_t *syncinfo,
			     int disable);
extern void md_enable_safemode(mddev_t *mddev, mdk_sync_support_t *syncinfo);
extern int md_init_sync_support(mddev_t *mddev, mdk_sync_support_t *syncinfo,
				const char *thread_string);
extern void md_teardown_sync_support(mddev_t *mddev,
				     mdk_sync_support_t *syncinfo);
extern mdk_work_handler_t md_sync_handler;
extern int md_get_sync_op_stat(mdk_sync_support_t *syncinfo,
	                  mdk_syncstatsop_arg_t *syncstat);
extern const char *md_member_type_string(mdk_member_t *member);

static inline void md_cancel_syncops(mdk_sync_support_t *syncinfo)
{
	md_cancel_workitems(&syncinfo->work_state);
}

static inline mdk_member_t *member_get(mdk_member_t *member)
{
        atomic_inc(&member->refcount);
	return (member);
}

static inline void member_put(mdk_member_t *member)
{
	if (!atomic_dec_and_test(&member->refcount))
		return;
	member->destructor(member);
}

static inline void member_reparent(mdk_member_t *member,
				   mdk_member_t *new_parent)
{
	md_member_orphan(member);

        if (new_parent)
		md_member_add_child(new_parent, member);
}

static inline void md_make_request(mdk_member_t *member, int rw,
				    struct buffer_head *bh)
{
	if (member->type == MDK_MTYPE_RDEV) {
		bh->b_dev = to_rdev(member)->kdev;
		bh->b_rdev = to_rdev(member)->kdev;
		generic_make_request(rw, bh);
	} else {
		(void)member->pers->make_request(member, rw, bh);
	}
}

static inline void md_update_work_support_string(mdk_workitem_support_t *work,
						 const char *fmt)
{
	if (fmt == NULL)
		fmt = "%s_work";
	md_update_thread_string(work->work_thread, fmt);
}

static inline void md_update_sync_support_string(mdk_sync_support_t *syncinfo,
						 const char *fmt)
{
	md_update_work_support_string(&syncinfo->work_state, fmt);
}

static inline int mddev_lock(mddev_t * mddev)
{
	return down_interruptible(&mddev->reconfig_sem);
}

static inline void mddev_lock_uninterruptible(mddev_t * mddev)
{
	down(&mddev->reconfig_sem);
}

static inline int mddev_trylock(mddev_t * mddev)
{
	return down_trylock(&mddev->reconfig_sem);
}

static inline void mddev_unlock(mddev_t * mddev)
{
	up(&mddev->reconfig_sem);

	if (mddev->thread)
		md_wakeup_thread(mddev->thread);
}


#define sector_div(n, b)( \
{ \
	int _res; \
	_res = (n) % (b); \
	(n) /= (b); \
	_res; \
} \
)

#define foreach_part(tmp,mddev)					\
	if (mddev->unit<MAX_MDP_DEVS)				\
		for(tmp = EMDMINOR(mddev);			\
		    tmp < EMDMINOR(mddev)+(1<<EMD_MINOR_SHIFT);	\
		    tmp++)

#define MD_BUG(x...) { printk("md: bug in file %s, line %d\n", __FILE__, __LINE__); md_print_devices(); }

static inline void md_clear_safemode_timer(mdk_sync_support_t *syncinfo)
{
	if (syncinfo->safemode == MDK_SAFEMODE_IDLE_WAIT) {
		del_timer(&syncinfo->safemode_timer);
		syncinfo->safemode = MDK_SAFEMODE_OFF;
	}
}

static inline void md_update_safemode_timer(mdk_sync_support_t *syncinfo)
{
	if (syncinfo->safemode != MDK_SAFEMODE_DISABLED) {
		mod_timer(&syncinfo->safemode_timer,
			  jiffies + syncinfo->safemode_delay);
		syncinfo->safemode = MDK_SAFEMODE_IDLE_WAIT;
	}
}

/*
 * Called before submitting normal I/O to a an array that
 * supports synchronization operations.  Blocks any write
 * requests that are inside the current sync lock range.
 * Returns the phase that should be assigned to this bio.
 *
 * XXX Prehaps too big to be inlined???
 */
static inline int syncable_vol_io_start(mdk_sync_support_t *syncinfo, int rw,
					struct buffer_head *bh,
					int *pbelow_sync_mark)
{
	sector_t bio_end;
	int phase;
	int below_sync_mark;

	bio_end = bh->b_rsector + (bh->b_size >> 9);
	spin_lock_irq(&syncinfo->sync_lock);
	if (rw != WRITE) {
		/*
		 * In the write case, the phase is not used.
		 * Return a value way out of range to catch
		 * improper use of this value.
		 */
		atomic_inc(&syncinfo->reads_pending);
		phase = INT_MAX;
	} else {
		wait_event_lock_irq(syncinfo->io_wait,
			    (bh->b_rsector >= syncinfo->lock_range.end
			  || bio_end < syncinfo->lock_range.start
			  || syncinfo->safemode == MDK_SAFEMODE_ACTIVE),
			    syncinfo->sync_lock);
		phase = syncinfo->phase;
		if (syncinfo->safemode == MDK_SAFEMODE_ACTIVE) {
			mddev_t *mddev;

			mddev = syncinfo->work_state.work_thread->mddev;
			md_exit_safemode(mddev, syncinfo, /*disable*/0);
		} else
			md_clear_safemode_timer(syncinfo);
		syncinfo->writes_pending[phase]++;
	}
	if (bio_end < syncinfo->lock_range.start)
		below_sync_mark = 1;
	else
		below_sync_mark = 0;
	spin_unlock_irq(&syncinfo->sync_lock);
	if (pbelow_sync_mark)
		*pbelow_sync_mark = below_sync_mark;
	return phase;
}

/*
 * Called on completion of normal I/O to a an array that
 * supports synchronization operations.  Notifies any operations
 * sleeping waiting for the current lock_range to clear of
 * pending requests.
 */
static inline void syncable_vol_io_complete(mdk_sync_support_t *syncinfo,
					    int rw, u_int bio_phase)
{
	if (rw != WRITE) {
		atomic_dec(&syncinfo->reads_pending);
		return;
	}
	spin_lock_irq(&syncinfo->sync_lock);
	syncinfo->writes_pending[bio_phase]--;
	if (syncinfo->writes_pending[bio_phase] == 0) {
		if (bio_phase != syncinfo->phase)
			wake_up_all(&syncinfo->sync_wait);
		if (syncinfo->writes_pending[syncinfo->phase] == 0)
			md_update_safemode_timer(syncinfo);
	}
	spin_unlock_irq(&syncinfo->sync_lock);
}

/*
 * Called when completing a sync I/O request.  If this is the
 * last I/O of a sync range, shrink the lock_range by the
 * sectors that are now synced out and wake any threads blocked
 * on that range.
 */
static inline int syncable_vol_syncio_start(mdk_sync_support_t *syncinfo)
{
	int phase;

	spin_lock_irq(&syncinfo->sync_lock);
	phase = syncinfo->phase;
	syncinfo->sync_pending[phase]++;
	spin_unlock_irq(&syncinfo->sync_lock);

	return phase;
}

/*
 * Called when completing a sync I/O request.  If this is the
 * last I/O of a sync range, shrink the lock_range by the
 * sectors that are now synced out and wake any threads blocked
 * on that range.
 */
static inline void syncable_vol_syncio_complete(mdk_sync_support_t *syncinfo,
						u_int phase)
{
	md_sector_range_t *sync_range;

	sync_range = &syncinfo->sync_range[phase];
	spin_lock_irq(&syncinfo->sync_lock);
	syncinfo->sync_pending[phase]--;
	if (syncinfo->sync_pending[phase] == 0
	 && (sync_range->start == sync_range->end)) {
		if (syncinfo->lock_range.start < sync_range->end)
			syncinfo->lock_range.start = sync_range->end;
		wake_up_all(&syncinfo->io_wait);
	} 
	spin_unlock_irq(&syncinfo->sync_lock);
}

/*
 * Unconditionally set a member's
 * state to the given state.
 */
static inline void md_set_member_state(mdk_member_t *member, mdu_mstate_t state)
{
	mdk_set_state_arg_t sstate;
	u_long flags;
	
	spin_lock_irqsave(&member->member_lock, flags);
	sstate.oldstate = member->state;
	spin_unlock_irqrestore(&member->member_lock, flags);
	sstate.newstate = state;
	member->metapers->member_op(member, MDK_MEMBER_OP_SET_STATE, &sstate);
}
#endif 

