#ifndef _RAID0_H
#define _RAID0_H

#include <linux/raid/emd/md.h>

struct strip_zone
{
	sector_t zone_offset;	/* Zone offset in md_dev */
	sector_t dev_offset;	/* Zone offset in real dev */
	sector_t size;		/* Zone size */
	int nb_dev;		/* # of devices attached to the zone */
	mdk_member_t **dev;	/* Devices attached to the zone */
};


typedef enum {
	RAID0_TF_NONE		= 0x00,
	RAID0_TF_PREALLOC	= 0x01
} raid0_trans_flags;

typedef enum {
	RAID0_OP_READ	= 0x00,
	RAID0_OP_WRITE	= 0x01
} raid0_optype;

struct raid0_transaction {
	struct buffer_head		*bh;
	raid0_trans_flags		flags;
	struct raid0_private_data	*conf;
	atomic_t			remaining;
	atomic_t			error_count;
	unsigned long			transaction_offset;
	raid0_optype			optype;
	struct raid0_transaction	*next_tbuf;
	struct raid0_transaction	*master_tbuf;
};

struct raid0_private_data
{
	struct strip_zone **hash_table; /* Table of indexes into strip_zone */
	struct strip_zone *strip_zone;
	mdk_member_t **devlist; /* lists of rdevs, pointed to by strip_zone->dev */
	int nr_strip_zones;
	mddev_t *mddev;
	sector_t hash_spacing;
	int preshift;			/* shift this before divide by hash_spacing */
	wait_queue_head_t wait_buffer;
	wait_queue_head_t wait_tbuf;
	mempool_t *r0_bh_pool;
	mempool_t *r0_tbuf_pool;
};

typedef struct raid0_private_data raid0_conf_t;

#define mddev_to_conf(mddev) ((raid0_conf_t *) mddev->member.pers_private)

#endif
